from __future__ import annotations

import logging
import os
import shutil
import tarfile
import tempfile
from os import PathLike
from pathlib import Path
from typing import IO, Any, Literal, Union
from urllib.parse import urlparse
from zipfile import ZipFile, is_zipfile

import requests
from requests.adapters import HTTPAdapter
from tqdm import tqdm
from urllib3.util.retry import Retry

logger = logging.getLogger(__name__)


OpenTextModeUpdating = Literal[
    "r+",
    "+r",
    "rt+",
    "r+t",
    "+rt",
    "tr+",
    "t+r",
    "+tr",
    "w+",
    "+w",
    "wt+",
    "w+t",
    "+wt",
    "tw+",
    "t+w",
    "+tw",
    "a+",
    "+a",
    "at+",
    "a+t",
    "+at",
    "ta+",
    "t+a",
    "+ta",
    "x+",
    "+x",
    "xt+",
    "x+t",
    "+xt",
    "tx+",
    "t+x",
    "+tx",
]
OpenTextModeWriting = Literal[
    "w",
    "wt",
    "tw",
    "a",
    "at",
    "ta",
    "x",
    "xt",
    "tx",
]
OpenTextModeReading = Literal[
    "r",
    "rt",
    "tr",
    "U",
    "rU",
    "Ur",
    "rtU",
    "rUt",
    "Urt",
    "trU",
    "tUr",
    "Utr",
]
OpenTextMode = Union[OpenTextModeUpdating, OpenTextModeWriting, OpenTextModeReading]
OpenBinaryModeUpdating = Literal[
    "rb+",
    "r+b",
    "+rb",
    "br+",
    "b+r",
    "+br",
    "wb+",
    "w+b",
    "+wb",
    "bw+",
    "b+w",
    "+bw",
    "ab+",
    "a+b",
    "+ab",
    "ba+",
    "b+a",
    "+ba",
    "xb+",
    "x+b",
    "+xb",
    "bx+",
    "b+x",
    "+bx",
]
OpenBinaryModeWriting = Literal[
    "wb",
    "bw",
    "ab",
    "ba",
    "xb",
    "bx",
]
OpenBinaryModeReading = Literal[
    "rb",
    "br",
    "rbU",
    "rUb",
    "Urb",
    "brU",
    "bUr",
    "Ubr",
]
OpenBinaryMode = Union[
    OpenBinaryModeUpdating, OpenBinaryModeReading, OpenBinaryModeWriting
]


def remove_file_or_directory(path: str | PathLike) -> None:
    try:
        path = Path(path)
        if path.is_dir():
            shutil.rmtree(path)
        else:
            os.remove(path)
    except FileNotFoundError:
        pass


def is_archive_file(filename: str | PathLike) -> bool:
    if not Path(filename).is_file():
        return False
    return is_zipfile(filename) or tarfile.is_tarfile(filename)


def extract_archive_file(
    source_path: str | PathLike,
    target_path: str | PathLike,
) -> None:
    with tempfile.TemporaryDirectory() as temp_dir:
        if is_zipfile(source_path):
            with ZipFile(source_path, "r") as zip_file:
                zip_file.extractall(temp_dir)
        elif tarfile.is_tarfile(source_path):
            with tarfile.open(source_path) as tar_file:
                tar_file.extractall(temp_dir)
        os.replace(temp_dir, target_path)


def extract_path(filename: str | PathLike) -> str:
    parsed = urlparse(str(filename))
    return parsed.path


def is_local(url_or_filename: str | PathLike) -> bool:
    if isinstance(url_or_filename, Path):
        return True

    parsed = urlparse(str(url_or_filename))
    if parsed.scheme in ("", "file", "osfs"):
        return True

    return False


def get_parent_path_and_filename(path: str | PathLike) -> tuple[str, str]:
    if isinstance(path, Path):
        parent = str(path.parent)
        name = str(path.name)
        return parent, name

    path = str(path)
    parsed_url = urlparse(path)
    scheme = parsed_url.scheme
    netloc = parsed_url.netloc
    path = parsed_url.path

    splitted = str(path).rsplit("/", 1)
    if len(splitted) == 2:
        parent, name = splitted
    else:
        parent = ""
        name = str(path)

    if scheme and netloc:
        parent = f"{scheme}://{netloc}/{parent}"

    return parent, name


def sizeof_fmt(num: int | float, suffix: str = "B") -> str:
    for unit in ["", "Ki", "Mi", "Gi", "Ti", "Pi", "Ei", "Zi"]:
        if abs(num) < 1024.0:
            return "%3.1f %s%s" % (num, unit, suffix)
        num /= 1024.0
    return "%.1f %s%s" % (num, "Yi", suffix)


def http_get(url: str, temp_file: IO[Any]) -> None:
    with _session_with_backoff() as session:
        req = session.get(url, stream=True)
        req.raise_for_status()
        content_length = req.headers.get("Content-Length")
        total = int(content_length) if content_length is not None else None
        with tqdm(
            unit="iB",
            unit_scale=True,
            unit_divisor=1024,
            total=total,
            desc="downloading",
        ) as progress:
            for chunk in req.iter_content(chunk_size=1024):
                if chunk:  # filter out keep-alive new chunks
                    progress.update(len(chunk))
                    temp_file.write(chunk)


def _session_with_backoff() -> requests.Session:
    session = requests.Session()
    retries = Retry(total=5, backoff_factor=1, status_forcelist=[502, 503, 504])
    session.mount("http://", HTTPAdapter(max_retries=retries))
    session.mount("https://", HTTPAdapter(max_retries=retries))

    return session



import logging
import os

from minato.commands import main

if os.environ.get("MINATO_DEBUG"):
    LEVEL = logging.DEBUG
else:
    level_name = os.environ.get("MINATO_LOG_LEVEL", "WARNING")
    LEVEL = logging._nameToLevel.get(level_name, logging.WARNING)

logging.basicConfig(
    format="%(asctime)s - %(levelname)s - %(name)s - %(message)s", level=LEVEL
)


def run() -> None:
    main(prog="minato")


if __name__ == "__main__":
    run()



from __future__ import annotations

import os
import sys
from typing import Any, TextIO


class Table:
    MIN_COLUMN_WIDTH = 2

    def __init__(self, columns: list[str], shrink: bool = True) -> None:
        self._columns = columns
        self._items: list[dict[str, Any]] = []
        self._shrink = shrink

    def __getitem__(self, columns: list[str]) -> "Table":
        table = Table(columns=columns)
        for item in self._items:
            table.add(item)
        return table

    @staticmethod
    def _get_padded_column_value(value: str, width: int) -> str:
        width = max(width, Table.MIN_COLUMN_WIDTH)
        if len(value) > width:
            value = value[: width - 1] + "\u2026"
        return f"{value:{width}}"

    @staticmethod
    def _get_column_value_str(value: Any) -> str:
        if isinstance(value, str):
            return repr(value)[1:-1]
        return repr(value)

    def _get_column_widths(self) -> dict[str, int]:
        column_widths: dict[str, int] = {}
        for col in self.columns:
            column_values = [x[col] for x in self._items]
            column_value_strings = [
                self._get_column_value_str(x) for x in column_values
            ]

            column_width = max(len(x) for x in column_value_strings + [col])

            column_widths[col] = column_width

        if not self._shrink:
            return column_widths

        num_columns = len(self.columns)
        terminal_width, _ = os.get_terminal_size()

        total_column_width = num_columns + sum(column_widths.values()) - 1
        previous_width = -1
        while (
            total_column_width > terminal_width and previous_width != total_column_width
        ):
            width, column = max(
                (width, column) for column, width in column_widths.items()
            )
            column_widths[column] = max(width - 1, Table.MIN_COLUMN_WIDTH)
            previous_width = total_column_width
            total_column_width = num_columns + sum(column_widths.values()) - 1
        return column_widths

    @property
    def columns(self) -> list[str]:
        return self._columns

    def add(self, item: dict[str, Any]) -> None:
        self._items.append({col: item[col] for col in self.columns})

    def sort(self, column: str, desc: bool = False) -> None:
        def _key(item: dict[str, Any]) -> Any:
            return item[column]

        self._items = sorted(self._items, key=_key, reverse=desc)

    def filter(self, query: str | dict[str, str]) -> "Table":
        if isinstance(query, str):
            query = {col: query for col in self.columns}
        table = Table(columns=self.columns)
        for item in self._items:
            for c, q in query.items():
                if q in item[c]:
                    table.add(item)
                    break
        return table

    def show(self, output: TextIO | None = None) -> None:
        if output is None:
            output = sys.stdout

        column_widths = self._get_column_widths()

        output.write(
            " ".join(
                self._get_padded_column_value(col, column_widths[col])
                for col in self.columns
            )
            + "\n"
        )
        output.write(" ".join("=" * column_widths[col] for col in self.columns) + "\n")
        for item in self._items:
            output.write(
                " ".join(
                    self._get_padded_column_value(
                        self._get_column_value_str(item[col]), column_widths[col]
                    )
                    for col in self.columns
                )
                + "\n"
            )



from __future__ import annotations

import argparse

from minato import __version__
from minato.commands.cache import CacheCommand  # noqa: F401
from minato.commands.list import ListCommand  # noqa: F401
from minato.commands.remove import RemoveCommand  # noqa: F401
from minato.commands.subcommand import Subcommand
from minato.commands.update import UpdateCommand  # noqa: F401


def create_subcommand(prog: str | None = None) -> Subcommand:
    parser = argparse.ArgumentParser(usage="%(prog)s", prog=prog)
    parser.add_argument(
        "--version",
        action="version",
        version="%(prog)s " + __version__,
    )
    return Subcommand(parser)


def main(prog: str | None = None) -> None:
    app = create_subcommand(prog)
    app()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['minato', 'minato.commands', 'minato.filesystems']

package_data = \
{'': ['*']}

install_requires = \
['requests>=2.25.1,<3.0.0', 'tqdm>=4.60.0,<5.0.0']

extras_require = \
{'all': ['boto3>=1.17.97,<2.0.0', 'google-cloud-storage>=1.39.0,<2.0.0'],
 'gcs': ['google-cloud-storage>=1.39.0,<2.0.0'],
 's3': ['boto3>=1.17.97,<2.0.0']}

entry_points = \
{'console_scripts': ['minato = minato.__main__:run']}

setup_kwargs = {
    'name': 'minato',
    'version': '0.9.0',
    'description': 'Cached file system for online resources',
    'long_description': 'Minato\n======\n\n[![Actions Status](https://github.com/altescy/minato/workflows/CI/badge.svg)](https://github.com/altescy/minato/actions/workflows/ci.yml)\n[![Python version](https://img.shields.io/pypi/pyversions/minato)](https://github.com/altescy/minato)\n[![License](https://img.shields.io/github/license/altescy/minato)](https://github.com/altescy/minato/blob/master/LICENSE)\n[![pypi version](https://img.shields.io/pypi/v/minato)](https://pypi.org/project/minato/)\n\nCache & file system for online resources in Python\n\n\n## Features\n\nMinato enables you to:\n- Download & cache online recsources\n  - minato supports the following protocols: HTTP(S) / AWS S3 / Google Cloud Storage\n  - You can manage cached files via command line interface\n- Automatically update cached files based on ETag\n  - minato downloads new versions if available when you access cached files\n- Open online files super easily\n  - By using `minato.open`, you can read/write online resources like the built-in `open` method\n\n## Installation\n\n```\npip install minato[all]\n```\n\n## Usage\n\n### Python\n\n```python\nimport minato\n\n# Read / write files on online storage\nwith minato.open("s3://your_bucket/path/to/file", "w") as f:\n    f.write("Create a new file on AWS S3!")\n\n# Cache & manage online resources in local storage\nlocal_filename = minato.cached_path("http://example.com/path/to/archive.zip!inner/path/to/file")\n```\n\n### CLI\n\n```\n❯ poetry run minato --help\nusage: minato\n\npositional arguments:\n  {cache,list,remove,update}\n    cache               cache remote file and return cached local file path\n    list                show list of cached files\n    remove              remove cached files\n    update              update cached files\n\noptional arguments:\n  -h, --help            show this help message and exit\n  --version             show program\'s version number and exit\n```\n',
    'author': 'Yasuhiro Yamaguchi',
    'author_email': 'altescy@fastmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/altescy/minato',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<3.11',
}


setup(**setup_kwargs)



from __future__ import annotations

import os
import tempfile
from contextlib import contextmanager
from os import PathLike
from typing import IO, Any, BinaryIO, ContextManager, Iterator, TextIO, overload

import requests

from minato.filesystems.filesystem import FileSystem
from minato.util import OpenBinaryMode, OpenTextMode, _session_with_backoff, http_get


@FileSystem.register(["http", "https"])
class HttpFileSystem(FileSystem):
    def exists(self) -> bool:
        response = requests.head(self._url.raw, allow_redirects=True)
        status_code = response.status_code
        return status_code == 200

    def download(self, path: str | PathLike) -> None:
        if not self.exists():
            raise FileNotFoundError(self._url.raw)

        with open(path, "w+b") as fp:
            http_get(self._url.raw, fp)

    def delete(self) -> None:
        raise OSError("HttpFileSystem cannot delete files or directories.")

    def get_version(self) -> str | None:
        if not self.exists():
            raise FileNotFoundError(self._url.raw)

        with _session_with_backoff() as session:
            response = session.head(self._url.raw, allow_redirects=True)
        if response.status_code != 200:
            raise OSError(
                "HEAD request failed for url {} with status code {}".format(
                    self._url.raw, response.status_code
                )
            )
        return response.headers.get("ETag")

    @overload
    def open_file(
        self,
        mode: OpenTextMode = ...,
        buffering: int = ...,
        encoding: str | None = ...,
        errors: str | None = ...,
        newline: str | None = ...,
    ) -> ContextManager[TextIO]:
        ...

    @overload
    def open_file(
        self,
        mode: OpenBinaryMode,
        buffering: int = ...,
        encoding: str | None = ...,
        errors: str | None = ...,
        newline: str | None = ...,
    ) -> ContextManager[BinaryIO]:
        ...

    @overload
    def open_file(
        self,
        mode: str,
        buffering: int = ...,
        encoding: str | None = ...,
        errors: str | None = ...,
        newline: str | None = ...,
    ) -> ContextManager[IO[Any]]:
        ...

    def open_file(
        self,
        mode: str = "r",
        buffering: int = -1,
        encoding: str | None = None,
        errors: str | None = None,
        newline: str | None = None,
    ) -> ContextManager[IO[Any]]:
        @contextmanager
        def _open(
            mode: str,
            buffering: int,
            encoding: str | None,
            errors: str | None,
            newline: str | None,
        ) -> Iterator[IO[Any]]:
            if not self.exists():
                raise FileNotFoundError(self._url.raw)

            if "a" in mode or "w" in mode or "+" in mode or "x" in mode:
                raise ValueError("HttpFileSystem is not writable.")

            temp_file = tempfile.NamedTemporaryFile(delete=False)
            try:
                http_get(self._url.raw, temp_file)
                temp_file.close()
                with open(
                    temp_file.name,
                    mode=mode,
                    encoding=encoding,
                    buffering=buffering,
                    errors=errors,
                    newline=newline,
                ) as fp:
                    yield fp
            finally:
                os.remove(temp_file.name)

        return _open(mode, buffering, encoding, errors, newline)
